# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import unittest

from ActionTree import *


class PreviewTestCase(unittest.TestCase):
    def test_dependencies_and_labels_are_not_only_equal_but_same(self):
        bLabel = ("b",)
        a = Action("a")
        b = Action(bLabel)
        a.add_dependency(b)

        otherB, = a.dependencies
        self.assertIs(otherB, b)
        self.assertIs(otherB.label, bLabel)

    def test_simple_preview(self):
        a = Action("a")
        self.assertEqual(a.get_preview(), ["a"])

    def test_preview_twice(self):
        # There was a bug where a second call to get_possible_execution_order would return [] :-/
        a = Action("a")
        self.assertEqual(a.get_preview(), ["a"])
        self.assertEqual(a.get_preview(), ["a"])

    def test_typed_label(self):
        a = Action(("a", "curious", "label", 42))
        self.assertEqual(a.get_preview(), [("a", "curious", "label", 42)])

    def test_none_label(self):
        a = Action(None)
        self.assertEqual(a.get_preview(), [None])

    def test_deep_dependency(self):
        a = Action("a")
        b = Action("b")
        c = Action("c")
        d = Action("d")
        a.add_dependency(b)
        b.add_dependency(c)
        c.add_dependency(d)

        self.assertEqual(a.get_preview(), ["d", "c", "b", "a"])

    def test_deep_dependency_with_duplicated_label(self):
        a = Action("label")
        b = Action("label")
        c = Action("label")
        d = Action("label")
        a.add_dependency(b)
        b.add_dependency(c)
        c.add_dependency(d)

        self.assertEqual(a.get_preview(), ["label", "label", "label", "label"])

    def test_diamond_dependency(self):
        #     a
        #    / \
        #   b   c
        #    \ /
        #     d

        a = Action("a")
        b = Action("b")
        c = Action("c")
        d = Action("d")
        a.add_dependency(b)
        a.add_dependency(c)
        b.add_dependency(d)
        c.add_dependency(d)

        self.assertIn(a.get_preview(), [["d", "c", "b", "a"], ["d", "b", "c", "a"]])



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import unittest

try:
    import unittest.mock
except ImportError:
    import mock
    unittest.mock = mock



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
Stock actions are predefined common tasks (manipulating the filesystem, calling an external program, etc.)
They all specialize :class:`.Action`.
"""

from __future__ import division, absolute_import, print_function

import errno
import os
import shutil
import subprocess
import time
from functools import partial

from . import Action, ActionFromCallable


class NullAction(Action):
    """
    A stock action that does nothing.
    Useful as a placeholder for several dependencies.
    """
    def __init__(self):
        Action.__init__(self, None)

    def do_execute(self):
        pass


class CallSubprocess(Action):
    """
    A stock action that calls a subprocess.

    Arguments are forwarded exactly to :func:`subprocess.check_call`.
    """
    def __init__(self, args, **kwds):
        Action.__init__(self, " ".join(args))
        self.__args = args
        self.__kwds = kwds

    def do_execute(self):
        subprocess.check_call(self.__args, **self.__kwds)


class CreateDirectory(Action):
    """
    A stock action that creates a directory.
    No error will be raised if the directory already exists.
    If the directory to create is nested, intermediate directories will be created as well.

    :param str name: the directory to create, passed to :func:`os.makedirs`.
    """
    def __init__(self, name):
        Action.__init__(self, "mkdir {}".format(name))
        self.__name = name

    def do_execute(self):
        try:
            os.makedirs(self.__name)
        except OSError as e:
            if e.errno != errno.EEXIST or not os.path.isdir(self.__name):
                raise


class DeleteFile(Action):
    """
    A stock action that deletes a file.
    No error will be raise if the file doesn't exist.

    :param str name: the name of the file to delete, passed to :func:`os.unlink`.
    """
    def __init__(self, name):
        Action.__init__(self, "rm {}".format(name))
        self.__name = name

    def do_execute(self):
        try:
            os.unlink(self.__name)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise


class CopyFile(Action):
    """
    A stock action that copies a file. Arguments are passed to :func:`shutil.copy`.

    :param str src: the file to copy
    :param str dst: the destination
    """
    def __init__(self, src, dst):
        Action.__init__(self, "cp {} {}".format(src, dst))
        self.__src = src
        self.__dst = dst

    def do_execute(self):
        shutil.copy(self.__src, self.__dst)


class TouchFile(Action):
    """
    A stock action that touches a file.
    If the file already exists, its modification time will be modified.
    Else, it will be created, empty.

    Note that the containing directory must exist.
    You might want to ensure that by adding a :class:`CreateDirectory` as a dependency.

    :param str name: the name of the file to touch. Passed to :func:`open` and/or :func:`os.utime`.
    """

    def __init__(self, name):
        Action.__init__(self, "touch {}".format(name))
        self.__name = name

    def do_execute(self):
        open(self.__name, "ab").close()  # Create the file if needed
        os.utime(self.__name, None)  # Actually change its time


class Sleep(Action):
    """
    A stock action that sleeps for a certain duration.

    :param float secs: seconds to sleep, passed to :func:`time.sleep`.
    """
    def __init__(self, secs):
        Action.__init__(self, "sleep {}".format(secs))
        self.__secs = secs

    def do_execute(self):
        time.sleep(self.__secs)



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import unittest

from ActionTree import *


class ExecutionTestCase(unittest.TestCase):
    def setUp(self):
        self.calls = []

    def __create_mocked_action(self, name):
        mock = unittest.mock.Mock()
        mock.side_effect = lambda: self.calls.append(name)
        action = ActionFromCallable(mock, name)
        return action, mock

    def test_simple_execution(self):
        a, aMock = self.__create_mocked_action("a")

        report = execute(a)

        self.assertTrue(report.is_success)
        self.assertEqual(report.get_action_status(a).status, ExecutionReport.ActionStatus.Successful)

        aMock.assert_called_once_with()

        self.assertEqual(self.calls, ["a"])

    def test_many_dependencies(self):
        #     a
        #    /|\
        #   / | \
        #  b  c  d

        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        d, dMock = self.__create_mocked_action("d")
        a.add_dependency(b)
        a.add_dependency(c)
        a.add_dependency(d)

        execute(a)

        aMock.assert_called_once_with()
        bMock.assert_called_once_with()
        cMock.assert_called_once_with()
        dMock.assert_called_once_with()

        self.assertEqual(self.calls[3:], ["a"])
        self.assertEqual(sorted(self.calls[:3]), ["b", "c", "d"])

    def test_deep_dependencies(self):
        #  a
        #  |
        #  b
        #  |
        #  c
        #  |
        #  d
        #  |
        #  e
        #  |
        #  f

        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        d, dMock = self.__create_mocked_action("d")
        e, eMock = self.__create_mocked_action("e")
        f, fMock = self.__create_mocked_action("f")
        a.add_dependency(b)
        b.add_dependency(c)
        c.add_dependency(d)
        d.add_dependency(e)
        e.add_dependency(f)

        execute(a)

        aMock.assert_called_once_with()
        bMock.assert_called_once_with()
        cMock.assert_called_once_with()
        dMock.assert_called_once_with()
        eMock.assert_called_once_with()
        fMock.assert_called_once_with()

        self.assertEqual(self.calls, ["f", "e", "d", "c", "b", "a"])

    def test_diamond_dependencies(self):
        #     a
        #    / \
        #   b   c
        #    \ /
        #     d

        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        d, dMock = self.__create_mocked_action("d")
        a.add_dependency(b)
        a.add_dependency(c)
        b.add_dependency(d)
        c.add_dependency(d)

        execute(a)

        self.assertEqual(self.calls[0:1], ["d"])
        self.assertEqual(sorted(self.calls[1:3]), ["b", "c"])
        self.assertEqual(self.calls[3:], ["a"])

    def test_half_diamond_dependency(self):
        #     a
        #    /|
        #   b |
        #    \|
        #     d

        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        d, dMock = self.__create_mocked_action("d")
        a.add_dependency(b)
        a.add_dependency(d)
        b.add_dependency(d)

        execute(a)

        self.assertEqual(self.calls, ["d", "b", "a"])

    def test_two_deep_branches(self):
        #     a
        #    / \
        #   b   c
        #   |   |
        #   d   e

        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        d, dMock = self.__create_mocked_action("d")
        e, eMock = self.__create_mocked_action("e")
        a.add_dependency(b)
        a.add_dependency(c)
        b.add_dependency(d)
        c.add_dependency(e)

        execute(a)

        aMock.assert_called_once_with()
        bMock.assert_called_once_with()
        cMock.assert_called_once_with()
        dMock.assert_called_once_with()
        eMock.assert_called_once_with()

        self.assertIn(
            self.calls,
            [
                # Leaves first
                ["d", "e", "b", "c", "a"],
                ["e", "d", "b", "c", "a"],
                ["d", "e", "c", "b", "a"],
                ["e", "d", "c", "b", "a"],
                # Full branch first
                ["d", "b", "e", "c", "a"],
                ["e", "c", "d", "b", "a"],
                # Leave, then full branch
                ["e", "d", "b", "c", "a"],
                ["d", "e", "c", "b", "a"],
            ]
        )



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import threading
import time
import unittest

from ActionTree import *


class ExecuteMock:
    def __init__(self, lock, mock):
        self.__mock = mock
        self.__lock = lock

    def __call__(self):
        with self.__lock:
            self.__mock.begin()
        time.sleep(0.1)
        with self.__lock:
            self.__mock.end()


class MultiThreadedExecutionTestCase(unittest.TestCase):
    def setUp(self):
        self.calls = []
        self.lock = threading.Lock()

    def __create_mocked_action(self, name):
        mock = unittest.mock.Mock()
        mock.begin.side_effect = lambda: self.calls.append("B" + name)
        mock.end.side_effect = lambda: self.calls.append("E" + name)
        action = ActionFromCallable(ExecuteMock(self.lock, mock), name)
        return action, mock

    def test_many_dependencies(self):
        #     a
        #    /|\
        #   / | \
        #  b  c  d

        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        d, dMock = self.__create_mocked_action("d")
        a.add_dependency(b)
        a.add_dependency(c)
        a.add_dependency(d)

        execute(a, jobs=3)

        aMock.begin.assert_called_once_with()
        aMock.end.assert_called_once_with()
        bMock.begin.assert_called_once_with()
        bMock.end.assert_called_once_with()
        cMock.begin.assert_called_once_with()
        cMock.end.assert_called_once_with()
        dMock.begin.assert_called_once_with()
        dMock.end.assert_called_once_with()

        self.assertEqual(sorted(self.calls[0:3]), ["Bb", "Bc", "Bd"])
        self.assertEqual(sorted(self.calls[3:6]), ["Eb", "Ec", "Ed"])
        self.assertEqual(self.calls[6:], ["Ba", "Ea"])

    def test_deep_dependencies(self):
        #  a
        #  |
        #  b
        #  |
        #  c
        #  |
        #  d
        #  |
        #  e
        #  |
        #  f

        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        d, dMock = self.__create_mocked_action("d")
        e, eMock = self.__create_mocked_action("e")
        f, fMock = self.__create_mocked_action("f")
        a.add_dependency(b)
        b.add_dependency(c)
        c.add_dependency(d)
        d.add_dependency(e)
        e.add_dependency(f)

        execute(a, jobs=3)

        aMock.begin.assert_called_once_with()
        aMock.end.assert_called_once_with()
        bMock.begin.assert_called_once_with()
        bMock.end.assert_called_once_with()
        cMock.begin.assert_called_once_with()
        cMock.end.assert_called_once_with()
        dMock.begin.assert_called_once_with()
        dMock.end.assert_called_once_with()
        eMock.begin.assert_called_once_with()
        eMock.end.assert_called_once_with()
        fMock.begin.assert_called_once_with()
        fMock.end.assert_called_once_with()

        self.assertEqual(self.calls, ["Bf", "Ef", "Be", "Ee", "Bd", "Ed", "Bc", "Ec", "Bb", "Eb", "Ba", "Ea"])

    def test_diamond_dependencies(self):
        #     a
        #    / \
        #   b   c
        #    \ /
        #     d

        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        d, dMock = self.__create_mocked_action("d")
        a.add_dependency(b)
        a.add_dependency(c)
        b.add_dependency(d)
        c.add_dependency(d)

        execute(a, jobs=3)

        aMock.begin.assert_called_once_with()
        aMock.end.assert_called_once_with()
        bMock.begin.assert_called_once_with()
        bMock.end.assert_called_once_with()
        cMock.begin.assert_called_once_with()
        cMock.end.assert_called_once_with()
        dMock.begin.assert_called_once_with()
        dMock.end.assert_called_once_with()

        self.assertEqual(self.calls[0:2], ["Bd", "Ed"])
        self.assertEqual(sorted(self.calls[2:4]), ["Bb", "Bc"])
        self.assertEqual(sorted(self.calls[4:6]), ["Eb", "Ec"])
        self.assertEqual(self.calls[6:], ["Ba", "Ea"])

    def test_half_diamond_dependency(self):
        #     a
        #    /|
        #   b |
        #    \|
        #     d

        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        d, dMock = self.__create_mocked_action("d")
        a.add_dependency(b)
        a.add_dependency(d)
        b.add_dependency(d)

        execute(a, jobs=3)

        self.assertEqual(self.calls, ["Bd", "Ed", "Bb", "Eb", "Ba", "Ea"])

    def test_two_deep_branches(self):
        #     a
        #    / \
        #   b   c
        #   |   |
        #   d   e

        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        d, dMock = self.__create_mocked_action("d")
        e, eMock = self.__create_mocked_action("e")
        a.add_dependency(b)
        a.add_dependency(c)
        b.add_dependency(d)
        c.add_dependency(e)

        execute(a, jobs=3)

        aMock.begin.assert_called_once_with()
        aMock.end.assert_called_once_with()
        bMock.begin.assert_called_once_with()
        bMock.end.assert_called_once_with()
        cMock.begin.assert_called_once_with()
        cMock.end.assert_called_once_with()
        dMock.begin.assert_called_once_with()
        dMock.end.assert_called_once_with()
        eMock.begin.assert_called_once_with()
        eMock.end.assert_called_once_with()

        self.assertEqual(sorted(self.calls[0:2]), ["Bd", "Be"])
        self.assertEqual(sorted(self.calls[2:6]), ["Bb", "Bc", "Ed", "Ee"])
        self.assertEqual(sorted(self.calls[6:8]), ["Eb", "Ec"])
        self.assertEqual(self.calls[8:], ["Ba", "Ea"])

    def test_default_jobs_count(self):
        # Default jobs count (when jobs < 0) is at least 2 even on a single core machine
        #     a
        #    /|
        #   / |
        #  b  c

        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        a.add_dependency(b)
        a.add_dependency(c)

        execute(a, jobs=-1)

        aMock.begin.assert_called_once_with()
        aMock.end.assert_called_once_with()
        bMock.begin.assert_called_once_with()
        bMock.end.assert_called_once_with()
        cMock.begin.assert_called_once_with()
        cMock.end.assert_called_once_with()

        self.assertEqual(sorted(self.calls[0:2]), ["Bb", "Bc"])
        self.assertEqual(sorted(self.calls[2:4]), ["Eb", "Ec"])
        self.assertEqual(self.calls[4:], ["Ba", "Ea"])



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import errno
import unittest

from ActionTree.stock import *
from ActionTree import *


class PatchingTestCase(unittest.TestCase):
    def patch(self, *args, **kwds):
        patcher = unittest.mock.patch(*args, **kwds)
        patched = patcher.start()
        self.addCleanup(patcher.stop)
        return patched


class CreateDirectoryTestCase(PatchingTestCase):
    def setUp(self):
        self.makedirs = self.patch("os.makedirs")
        self.isdir = self.patch("os.path.isdir")

    def test_label(self):
        self.assertEqual(CreateDirectory("xxx").label, "mkdir xxx")

        self.makedirs.assert_not_called()
        self.isdir.assert_not_called()

    def test_success(self):
        self.makedirs.expect("xxx")

        execute(CreateDirectory("xxx"))

        self.makedirs.assert_called_once_with("xxx")
        self.isdir.assert_not_called()

    def test_directory_exists(self):
        self.makedirs.side_effect = OSError(errno.EEXIST, "File exists")
        self.isdir.return_value = True

        execute(CreateDirectory("xxx"))

        self.makedirs.assert_called_once_with("xxx")
        self.isdir.assert_called_once_with("xxx")

    def test_file_exists(self):
        self.makedirs.side_effect = OSError(errno.EEXIST, "File exists")
        self.isdir.return_value = False

        with self.assertRaises(CompoundException):
            execute(CreateDirectory("xxx"))

        self.makedirs.assert_called_once_with("xxx")
        self.isdir.assert_called_once_with("xxx")

    def test_other_failure(self):
        self.makedirs.side_effect = OSError(-1, "Foobar")

        with self.assertRaises(CompoundException):
            execute(CreateDirectory("xxx"))

        self.makedirs.assert_called_once_with("xxx")
        self.isdir.assert_not_called()


class CallSubprocessTestCase(PatchingTestCase):
    def setUp(self):
        self.check_call = self.patch("subprocess.check_call")

    def test_label(self):
        self.assertEqual(CallSubprocess(["xxx", "yyy"]).label, "xxx yyy")

        self.check_call.assert_not_called()

    def test_simple_call(self):
        execute(CallSubprocess(["xxx"]))

        self.check_call.assert_called_once_with(["xxx"])

    def test_call_with_several_args(self):
        self.check_call.expect(["xxx", "yyy"])
        execute(CallSubprocess(["xxx", "yyy"]))

        self.check_call.assert_called_once_with(["xxx", "yyy"])

    def test_call_with_kwds(self):
        execute(CallSubprocess(["xxx", "yyy"], foo="bar"))

        self.check_call.assert_called_once_with(["xxx", "yyy"], foo="bar")


class DeleteFileTestCase(PatchingTestCase):
    def setUp(self):
        self.unlink = self.patch("os.unlink")

    def test_label(self):
        self.assertEqual(DeleteFile("xxx").label, "rm xxx")

        self.unlink.assert_not_called()

    def test_success(self):
        execute(DeleteFile("xxx"))

        self.unlink.assert_called_once_with("xxx")

    def test_file_does_not_exist(self):
        self.unlink.side_effect = OSError(errno.ENOENT, "No such file or directory")

        execute(DeleteFile("xxx"))

        self.unlink.assert_called_once_with("xxx")

    def test_other_failure(self):
        self.unlink.side_effect = OSError(-1, "Foobar")

        with self.assertRaises(CompoundException):
            execute(DeleteFile("xxx"))

        self.unlink.assert_called_once_with("xxx")


class CopyFileTestCase(PatchingTestCase):
    def setUp(self):
        self.copy = self.patch("shutil.copy")

    def test_label(self):
        self.assertEqual(CopyFile("from", "to").label, "cp from to")

        self.copy.assert_not_called()

    def test_success(self):
        execute(CopyFile("from", "to"))

        self.copy.assert_called_once_with("from", "to")

    def test_failure(self):
        self.copy.side_effect = OSError(-1, "Foobar")

        with self.assertRaises(CompoundException):
            execute(CopyFile("from", "to"))

        self.copy.assert_called_once_with("from", "to")


class TouchFileTestCase(PatchingTestCase):
    def setUp(self):
        self.open = self.patch("ActionTree.stock.open", new=unittest.mock.mock_open(), create=True)
        self.utime = self.patch("os.utime")

    def test_label(self):
        self.assertEqual(TouchFile("xxx").label, "touch xxx")

        self.open.assert_not_called()
        self.utime.assert_not_called()

    def test_success(self):
        execute(TouchFile("xxx"))

        self.open.assert_called_once_with("xxx", "ab")
        self.open().close.assert_called_once_with()
        self.utime.assert_called_once_with("xxx", None)


class NullActionTestCase(unittest.TestCase):
    def test(self):
        execute(NullAction())


class SleepTestCase(PatchingTestCase):
    def setUp(self):
        self.sleep = self.patch("time.sleep")

    def test(self):
        execute(Sleep(1))

        self.sleep.assert_called_once_with(1)



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import unittest

from ActionTree import *


class Counter:
    def __init__(self):
        self.__value = 0

    def reset(self):
        self.__value = 0

    def __call__(self):
        self.__value += 1
        return self.__value


class TimingTestCase(unittest.TestCase):
    def __create_mocked_action(self, name):
        mock = unittest.mock.Mock()
        action = ActionFromCallable(mock, name)
        return action, mock

    def setUp(self):
        patcher = unittest.mock.patch("datetime.datetime")
        self.datetime = patcher.start()
        self.datetime.now.side_effect = Counter()
        self.addCleanup(patcher.stop)

    def test_success(self):
        a, aMock = self.__create_mocked_action("a")

        report = execute(a)

        self.assertEqual(report.get_action_status(a).ready_time, 1)
        self.assertIsNone(report.get_action_status(a).cancel_time)
        self.assertEqual(report.get_action_status(a).start_time, 2)
        self.assertEqual(report.get_action_status(a).success_time, 3)
        self.assertIsNone(report.get_action_status(a).failure_time)

    def test_failure(self):
        a, aMock = self.__create_mocked_action("a")
        aMock.side_effect = Exception()

        with self.assertRaises(CompoundException) as catcher:
            execute(a)
        report = catcher.exception.execution_report

        self.assertEqual(report.get_action_status(a).ready_time, 1)
        self.assertIsNone(report.get_action_status(a).cancel_time)
        self.assertEqual(report.get_action_status(a).start_time, 2)
        self.assertIsNone(report.get_action_status(a).success_time)
        self.assertEqual(report.get_action_status(a).failure_time, 3)

    def test_cancelation_before_ready(self):
        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        bMock.side_effect = Exception()
        a.add_dependency(b)

        with self.assertRaises(CompoundException) as catcher:
            execute(a)
        report = catcher.exception.execution_report

        self.assertEqual(report.get_action_status(b).ready_time, 1)
        self.assertIsNone(report.get_action_status(b).cancel_time)
        self.assertEqual(report.get_action_status(b).start_time, 2)
        self.assertIsNone(report.get_action_status(b).success_time)
        self.assertEqual(report.get_action_status(b).failure_time, 3)

        self.assertIsNone(report.get_action_status(a).ready_time)
        self.assertEqual(report.get_action_status(a).cancel_time, 4)
        self.assertIsNone(report.get_action_status(a).start_time)
        self.assertIsNone(report.get_action_status(a).success_time)
        self.assertIsNone(report.get_action_status(a).failure_time)

    def test_cancelation_with_keep_going(self):
        for i in range(10):
            self.datetime.now.side_effect.reset()

            a0, a0Mock = self.__create_mocked_action("a0")
            a, aMock = self.__create_mocked_action("a")
            a0.add_dependency(a)
            b, bMock = self.__create_mocked_action("b")
            bMock.side_effect = Exception()
            a.add_dependency(b)
            DEPS = 10
            deps = []
            for i in range(DEPS):
                c, cMock = self.__create_mocked_action("c")
                a.add_dependency(c)
                deps.append(c)

            with self.assertRaises(CompoundException) as catcher:
                execute(a0, keep_going=True)
            report = catcher.exception.execution_report

            # a is not cancelled before all its dependencies are done
            self.assertGreater(report.get_action_status(a).cancel_time, 3 + 2 * DEPS)
            # a0 is cancelled at the same time as a
            self.assertEqual(report.get_action_status(a0).cancel_time, report.get_action_status(a).cancel_time)



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import textwrap
import unittest

from ActionTree import *


class GraphTestCase(unittest.TestCase):
    def test_internal_graph_not_returned(self):
        g = DependencyGraph(ActionFromCallable(None, "a"))
        g1 = g.get_graphviz_graph()
        self.assertEqual(g1.format, "pdf")
        g1.format = "png"
        self.assertEqual(g.get_graphviz_graph().format, "pdf")

    def test_single_action(self):
        a = ActionFromCallable(None, "a")

        self.assertEqual(
            DependencyGraph(a).get_graphviz_graph().source,
            textwrap.dedent(
                """\
                digraph action {
                \tnode [shape=box]
                \t0 [label=a]
                }"""
            )
        )

    def test_dependency(self):
        b = ActionFromCallable(None, "b")
        a = ActionFromCallable(None, "a")
        a.add_dependency(b)

        self.assertEqual(
            DependencyGraph(a).get_graphviz_graph().source,
            textwrap.dedent(
                """\
                digraph action {
                \tnode [shape=box]
                \t0 [label=b]
                \t1 [label=a]
                \t\t1 -> 0
                }"""
            )
        )

    def test_add_dependency_after_constructing_graph(self):
        a = ActionFromCallable(None, "a")
        g = DependencyGraph(a)
        a.add_dependency(ActionFromCallable(None, "b"))

        self.assertEqual(
            g.get_graphviz_graph().source,
            textwrap.dedent(
                """\
                digraph action {
                \tnode [shape=box]
                \t0 [label=a]
                }"""
            )
        )

    def test_diamond(self):
        for i in range(1000):
            a = ActionFromCallable(None, "a")
            b = ActionFromCallable(None, "b")
            b.add_dependency(a)
            c = ActionFromCallable(None, "c")
            c.add_dependency(a)
            d = ActionFromCallable(None, "d")
            d.add_dependency(c)
            d.add_dependency(b)

            self.assertIn(
                DependencyGraph(d).get_graphviz_graph().source,
                [
                    textwrap.dedent(
                        """\
                        digraph action {
                        \tnode [shape=box]
                        \t0 [label=a]
                        \t1 [label=b]
                        \t\t1 -> 0
                        \t2 [label=c]
                        \t\t2 -> 0
                        \t3 [label=d]
                        \t\t3 -> 1
                        \t\t3 -> 2
                        }"""
                    ),
                    textwrap.dedent(
                        """\
                        digraph action {
                        \tnode [shape=box]
                        \t0 [label=a]
                        \t1 [label=c]
                        \t\t1 -> 0
                        \t2 [label=b]
                        \t\t2 -> 0
                        \t3 [label=d]
                        \t\t3 -> 1
                        \t\t3 -> 2
                        }"""
                    ),
                ]
            )

    def test_typed_label(self):
        a = ActionFromCallable(None, ("a", "curious", "label", 42))

        self.assertEqual(
            DependencyGraph(a).get_graphviz_graph().source,
            textwrap.dedent(
                """\
                digraph action {
                \tnode [shape=box]
                \t0 [label="('a', 'curious', 'label', 42)"]
                }"""
            )
        )

    def test_weird_string_label(self):
        a = ActionFromCallable(None, "spaces and; semi=columns")

        self.assertEqual(
            DependencyGraph(a).get_graphviz_graph().source,
            textwrap.dedent(
                """\
                digraph action {
                \tnode [shape=box]
                \t0 [label="spaces and; semi=columns"]
                }"""
            )
        )

    def test_None_label(self):
        a = ActionFromCallable(None, None)

        self.assertEqual(
            DependencyGraph(a).get_graphviz_graph().source,
            textwrap.dedent(
                """\
                digraph action {
                \tnode [shape=box]
                \t0 [label=None]
                }"""
            )
        )

    def test_None_label_twice(self):
        a = ActionFromCallable(None, None)
        b = ActionFromCallable(None, None)
        a.add_dependency(b)

        self.assertEqual(
            DependencyGraph(a).get_graphviz_graph().source,
            textwrap.dedent(
                """\
                digraph action {
                \tnode [shape=box]
                \t0 [label=None]
                \t1 [label=None]
                \t\t1 -> 0
                }"""
            )
        )



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

# import datetime
# import functools
import unittest

from ActionTree import *


class UtilitiesTestCase(unittest.TestCase):
    def test_nearest_before_first(self):
        self.assertEqual(GanttChart._GanttChart__nearest(2, [10, 20, 30]), 10)

    def test_nearest_after_last(self):
        self.assertEqual(GanttChart._GanttChart__nearest(35, [10, 20, 30]), 30)

    def test_nearest_in_the_middle(self):
        self.assertEqual(GanttChart._GanttChart__nearest(18, [10, 20, 30]), 20)
        self.assertEqual(GanttChart._GanttChart__nearest(22, [10, 20, 30]), 20)


class Counter:
    def __init__(self):
        self.__value = datetime.datetime(2017, 6, 8, 13, 47, 12)

    def __call__(self):
        self.__value += datetime.timedelta(seconds=2)
        return self.__value


class GanttChartTestCase(unittest.TestCase):
    def __create_mocked_action(self, name):
        mock = unittest.mock.Mock()
        action = ActionFromCallable(mock, name)
        return action, mock

    def setUp(self):
        counter = Counter()
        patcher = unittest.mock.patch("datetime.datetime")
        self.datetime = patcher.start()
        self.datetime.now.side_effect = counter
        self.addCleanup(patcher.stop)

    def test(self):
        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        c.add_dependency(a)
        c.add_dependency(b)

        chart = GanttChart(execute(c, jobs=2))
        ax = unittest.mock.Mock()
        chart.plot_on_mpl_axes(ax)
        # @todo Assert something about calls to ax



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import concurrent.futures as futures
import datetime
import multiprocessing
import os.path

import graphviz
import matplotlib
import matplotlib.dates
import matplotlib.figure
import matplotlib.backends.backend_agg


def execute(action, jobs=1, keep_going=False):
    """
    Recursively execute an action's dependencies then the action.

    :param Action action: the action to execute.
    :param int jobs: number of actions to execute in parallel. Pass ``None`` to let ActionTree choose.
    :param bool keep_going: if ``True``, then execution does not stop on first failure,
        but executes as many dependencies as possible.

    :raises CompoundException: when dependencies raise exceptions.

    :rtype: ExecutionReport
    """
    if jobs <= 0 or jobs is None:
        jobs = multiprocessing.cpu_count() + 1
    return Executor(jobs, keep_going).execute(action)


class ExecutionReport(object):
    """
    ExecutionReport()

    Execution report, returned by :func:`.execute`.
    """

    class ActionStatus(object):
        """
        ActionStatus()

        Status of a single :class:`.Action`.
        """

        Successful = "Successful"
        "The :attr:`status` after a successful execution."
        Canceled = "Canceled"
        "The :attr:`status` after a failed execution where a dependency raised an exception."
        Failed = "Failed"
        "The :attr:`status` after a failed execution where this action raised an exception."

        def __init__(self, ready_time=None, start_time=None, cancel_time=None, failure_time=None, success_time=None):
            ready = bool(ready_time)
            start = bool(start_time)
            cancel = bool(cancel_time)
            failure = bool(failure_time)
            success = bool(success_time)
            if ready:
                if start:
                    assert not cancel
                    assert (failure or success)
                    self.__status = self.Successful if success else self.Failed
                else:
                    assert cancel
                    assert not (failure or success)
                    self.__status = self.Canceled
            else:
                assert cancel
                assert not (start or failure or success)
                self.__status = self.Canceled
            self.__ready_time = ready_time
            self.__cancel_time = cancel_time
            self.__start_time = start_time
            self.__success_time = success_time
            self.__failure_time = failure_time

        @property
        def status(self):
            """
            @todo Document
            """
            return self.__status

        @property
        def ready_time(self):
            """
            The local :class:`~datetime.datetime` when this action was ready to execute.
            """
            return self.__ready_time

        @property
        def start_time(self):
            """
            The local :class:`~datetime.datetime` at the begining of the execution of this action.
            """
            return self.__start_time

        @property
        def success_time(self):
            """
            The local :class:`~datetime.datetime` at the successful end of the execution of this action.
            """
            return self.__success_time

        @property
        def failure_time(self):
            """
            The local :class:`~datetime.datetime` at the successful end of the execution of this action.
            """
            return self.__failure_time

        @property
        def cancel_time(self):
            """
            The local :class:`~datetime.datetime` when this action was canceled.
            """
            return self.__cancel_time

    def __init__(self):
        self.__is_success = True
        self.__action_statuses = dict()

    def set_success(self, is_success):
        self.__is_success = is_success

    @property
    def is_success(self):
        """
        ``True`` if the execution finished without error.

        :rtype: bool
        """
        return self.__is_success

    def set_action_status(self, action, status):
        self.__action_statuses[action] = status

    def get_action_status(self, action):
        """
        @todo Document
        """
        return self.__action_statuses[action]

    def get_actions_and_statuses(self):
        """
        @todo Document
        """
        return self.__action_statuses.items()


class Executor(object):
    def __init__(self, jobs, keep_going):
        self.__jobs = jobs
        self.__keep_going = keep_going

    class Execution:
        # An aggregate for things that cannot be stored in Executor
        # (to allow several parallel calls to Executor.execute)
        def __init__(self, executor, pending):
            self.executor = executor
            self.pending = set(pending)  # Action
            self.submitted = dict()  # Future -> Action
            self.submitted_at = dict()  # Action -> datetime.datetime
            self.succeeded = set()  # Action
            self.failed = set()  # Action
            self.exceptions = []
            self.report = ExecutionReport()

    def execute(self, action):
        # Threads in pool just call self.__time_execute, which has no side effects.
        # To avoid races, only the thread calling Executor.execute is allowed to modify anything.

        with futures.ThreadPoolExecutor(max_workers=self.__jobs) as executor:
            execution = Executor.Execution(executor, action.get_all_dependencies())
            while execution.pending or execution.submitted:
                self.__progress(execution)

        if execution.exceptions:
            raise CompoundException(execution.exceptions, execution.report)
        else:
            return execution.report

    def __progress(self, execution):
        now = datetime.datetime.now()
        if self.__keep_going or not execution.exceptions:
            self.__submit_or_cancel(execution, now)
        else:
            self.__cancel(execution, now)
        self.__wait(execution)

    def __submit_or_cancel(self, execution, now):
        go_on = True
        while go_on:
            go_on = False
            for action in set(execution.pending):
                done = execution.succeeded | execution.failed
                if all(d in done for d in action.dependencies):
                    if any(d in execution.failed for d in action.dependencies):
                        self.__mark_action_canceled(execution, action, now)
                        execution.pending.remove(action)
                        go_on = True
                    else:
                        execution.submitted[execution.executor.submit(self.__time_execute, action)] = action
                        execution.submitted_at[action] = now
                        execution.pending.remove(action)
                        go_on = True

    @staticmethod
    def __time_execute(action):
        exception = None
        try:
            begin_time = datetime.datetime.now()
            action.do_execute()
        except Exception as e:
            exception = e
        end_time = datetime.datetime.now()
        return (exception, begin_time, end_time)

    def __cancel(self, execution, now):
        for (f, action) in execution.submitted.items():
            if f.cancel():
                self.__mark_action_canceled(execution, action, now)
                del execution.submitted[f]
        for action in execution.pending:
            self.__mark_action_canceled(execution, action, now)
        execution.pending.clear()

    def __wait(self, execution):
        waited = futures.wait(execution.submitted.keys(), return_when=futures.FIRST_COMPLETED)
        for f in waited.done:
            action = execution.submitted[f]
            del execution.submitted[f]
            (exception, begin_time, end_time) = f.result()
            if exception:
                self.__mark_action_failed(execution, action, begin_time, end_time)
                execution.exceptions.append(exception)
                execution.report.set_success(False)
            else:
                self.__mark_action_successful(execution, action, begin_time, end_time)

    @staticmethod
    def __mark_action_canceled(execution, action, cancel_time):
        execution.failed.add(action)
        execution.report.set_action_status(
            action,
            ExecutionReport.ActionStatus(
                ready_time=execution.submitted_at.get(action),
                cancel_time=cancel_time,
            )
        )

    @staticmethod
    def __mark_action_successful(execution, action, start_time, success_time):
        execution.succeeded.add(action)
        execution.report.set_action_status(
            action,
            ExecutionReport.ActionStatus(
                ready_time=execution.submitted_at[action],
                start_time=start_time,
                success_time=success_time,
            ),
        )

    @staticmethod
    def __mark_action_failed(execution, action, start_time, failure_time):
        execution.failed.add(action)
        execution.report.set_action_status(
            action,
            ExecutionReport.ActionStatus(
                ready_time=execution.submitted_at[action],
                start_time=start_time,
                failure_time=failure_time,
            )
        )


class Action(object):
    """
    The main class of ActionTree.
    An action to be started after all its dependencies are finished.
    Pass it to :func:`.execute`.

    This is a base class for your custom actions.
    You must define a ``def do_execute(self):`` method that performs the action.
    Its return value is ignored.
    If it raises and exception, it is captured and re-raised in a :exc:`CompoundException`.

    See also :class:`.ActionFromCallable` if you just want to create an action from a simple callable.
    """
    # @todo Add a note about printing anything in do_execute
    # @todo Add a note saying

    def __init__(self, label):
        """
        :param label: whatever you want to attach to the action.
            Can be retrieved by :attr:`label` and :meth:`get_preview`.
        """
        self.__label = label
        self.__dependencies = set()

    @property
    def label(self):
        """
        The label passed to the constructor.
        """
        return self.__label

    def add_dependency(self, dependency):
        """
        Add a dependency to be executed before this action.
        Order of insertion of dependencies is not important.

        :param Action dependency:

        :raises DependencyCycleException: when adding the new dependency would create a cycle.
        """
        if self in dependency.get_all_dependencies():
            raise DependencyCycleException()
        self.__dependencies.add(dependency)

    @property
    def dependencies(self):
        """
        Return the list of this action's dependencies.
        """
        return list(self.__dependencies)

    def get_all_dependencies(self):
        """
        Return the set of this action's recursive dependencies, including itself.
        """
        dependencies = set([self])
        for dependency in self.__dependencies:
            dependencies |= dependency.get_all_dependencies()
        return dependencies

    def get_preview(self):
        """
        Return the labels of this action and its dependencies, in an order that could be the execution order.
        """
        return [action.__label for action in self.get_possible_execution_order()]

    def get_possible_execution_order(self, seen_actions=None):
        if seen_actions is None:
            seen_actions = set()
        actions = []
        if self not in seen_actions:
            seen_actions.add(self)
            for dependency in self.__dependencies:
                actions += dependency.get_possible_execution_order(seen_actions)
            actions.append(self)
        return actions


class ActionFromCallable(Action):
    """
    An :class:`.Action` sub-class for the simple use-case of using a plain callable as an action.
    """

    def __init__(self, do_execute, label):
        """
        :param label: see :class:`.Action`.
        :param callable do_execute: the function to execute the action.
        """
        super(ActionFromCallable, self).__init__(label)
        self.__do_execute = do_execute

    def do_execute(self):
        self.__do_execute()


class CompoundException(Exception):
    """
    Exception thrown by :func:`.execute` when a dependencies raise exceptions.
    """

    def __init__(self, exceptions, execution_report):
        super(CompoundException, self).__init__(exceptions)
        self.__exceptions = exceptions
        self.__execution_report = execution_report

    @property
    def exceptions(self):
        """
        The list of the encapsulated exceptions.
        """
        return self.__exceptions

    @property
    def execution_report(self):
        """
        The :class:`.ExecutionReport` of the failed execution.
        """
        return self.__execution_report


class DependencyCycleException(Exception):
    """
    Exception thrown by :meth:`.Action.add_dependency` when adding the new dependency would create a cycle.
    """

    def __init__(self):
        super(DependencyCycleException, self).__init__("Dependency cycle")


class DependencyGraph(object):
    """
    @todo Document
    """

    def __init__(self, action):
        self.__graphviz_graph = graphviz.Digraph("action", node_attr={"shape": "box"})
        nodes = {}
        for (i, action) in enumerate(action.get_possible_execution_order()):
            node = str(i)
            nodes[action] = node
            self.__graphviz_graph.node(node, str(action.label))
            for dependency in action.dependencies:
                assert dependency in nodes  # Because we are iterating a possible execution order
                self.__graphviz_graph.edge(node, nodes[dependency])

    def write_to_png(self, filename):  # pragma no cover (Untestable? But small.)
        """
        Write the graph as a PNG image to the specified file.

        See also :meth:`get_graphviz_graph` if you want to draw the graph somewhere else.
        """
        directory = os.path.dirname(filename)
        filename = os.path.basename(filename)
        filename, ext = os.path.splitext(filename)
        g = self.get_graphviz_graph()
        g.format = "png"
        g.render(directory=directory, filename=filename, cleanup=True)

    def get_graphviz_graph(self):
        """
        Return a :class:`graphviz.Digraph` of this dependency graph.

        See also :meth:`write_to_png` for the simplest use-case.
        """
        return self.__graphviz_graph.copy()


class GanttChart(object):  # pragma no cover (Too difficult to unit test)
    """
    @todo Document
    """

    def __init__(self, report):
        # @todo Sort actions somehow to improve readability (top-left to bottom-right)
        self.__actions = {
            id(action): self.__make_action(action, status)
            for (action, status) in report.get_actions_and_statuses()
        }

    # @todo Factorize Actions
    class SuccessfulAction(object):
        def __init__(self, action, status):
            self.__label = str(action.label)
            self.__id = id(action)
            self.__dependencies = set(id(d) for d in action.dependencies)
            self.__ready_time = status.ready_time
            self.__start_time = status.start_time
            self.__success_time = status.success_time

        @property
        def min_time(self):
            return self.__ready_time

        @property
        def max_time(self):
            return self.__success_time

        def draw(self, ax, ordinates, actions):
            ordinate = ordinates[self.__id]
            ax.plot([self.__ready_time, self.__start_time], [ordinate, ordinate], color="blue", lw=1)
            # @todo Use an other end-style to avoid pixels before/after min/max_time
            ax.plot([self.__start_time, self.__success_time], [ordinate, ordinate], color="blue", lw=4)
            # @todo Make sure the text is not outside the plot on the right
            ax.annotate(self.__label, xy=(self.__start_time, ordinate), xytext=(0, 3), textcoords="offset points")
            for d in self.__dependencies:
                ax.plot([actions[d].max_time, self.min_time], [ordinates[d], ordinate], "k:", lw=1)

    class FailedAction(object):
        def __init__(self, action, status):
            self.__label = str(action.label)
            self.__id = id(action)
            self.__dependencies = set(id(d) for d in action.dependencies)
            self.__ready_time = status.ready_time
            self.__start_time = status.start_time
            self.__failure_time = status.failure_time

        @property
        def min_time(self):
            return self.__ready_time

        @property
        def max_time(self):
            return self.__failure_time

        def draw(self, ax, ordinates, actions):
            ordinate = ordinates[self.__id]
            ax.plot([self.__ready_time, self.__start_time], [ordinate, ordinate], color="red", lw=1)
            ax.plot([self.__start_time, self.__failure_time], [ordinate, ordinate], color="red", lw=4)
            ax.annotate(self.__label, xy=(self.__start_time, ordinate), xytext=(0, 3), textcoords="offset points")
            for d in self.__dependencies:
                ax.plot([actions[d].max_time, self.min_time], [ordinates[d], ordinate], "k:", lw=1)

    class CanceledAction(object):
        def __init__(self, action, status):
            self.__label = str(action.label)
            self.__id = id(action)
            self.__dependencies = set(id(d) for d in action.dependencies)
            self.__ready_time = status.ready_time
            self.__cancel_time = status.cancel_time

        @property
        def min_time(self):
            return self.__cancel_time if self.__ready_time is None else self.__ready_time

        @property
        def max_time(self):
            return self.__cancel_time

        def draw(self, ax, ordinates, actions):
            ordinate = ordinates[self.__id]
            if self.__ready_time:
                ax.plot([self.__ready_time, self.__cancel_time], [ordinate, ordinate], color="grey", lw=1)
            ax.annotate(
                "(Canceled) {}".format(self.__label),
                xy=(self.__cancel_time, ordinate),
                xytext=(0, 3),
                textcoords="offset points"
            )
            for d in self.__dependencies:
                ax.plot([actions[d].max_time, self.min_time], [ordinates[d], ordinate], "k:", lw=1)

    @classmethod
    def __make_action(cls, action, status):
        if status.status == ExecutionReport.ActionStatus.Successful:
            return cls.SuccessfulAction(action, status)
        elif status.status == ExecutionReport.ActionStatus.Failed:
            return cls.FailedAction(action, status)
        elif status.status == ExecutionReport.ActionStatus.Canceled:
            return cls.CanceledAction(action, status)

    def write_to_png(self, filename):
        """
        Write the Gantt chart as a PNG image to the specified file.

        See also :meth:`get_mpl_figure` and :meth:`plot_on_mpl_axes` if you want to draw the report somewhere else.
        """
        figure = self.get_mpl_figure()
        canvas = matplotlib.backends.backend_agg.FigureCanvasAgg(figure)
        canvas.print_figure(filename)

    def get_mpl_figure(self):
        """
        Return a :class:`matplotlib.figure.Figure` of this Gantt chart.

        See also :meth:`plot_on_mpl_axes` if you want to draw the Gantt chart on your own matplotlib figure.

        See also :meth:`write_to_png` for the simplest use-case.
        """
        fig = matplotlib.figure.Figure()
        ax = fig.add_subplot(1, 1, 1)

        self.plot_on_mpl_axes(ax)

        return fig

    @staticmethod
    def __nearest(v, values):
        for i, value in enumerate(values):
            if v < value:
                break
        if i == 0:
            return values[0]
        else:
            if v - values[i - 1] <= values[i] - v:
                return values[i - 1]
            else:
                return values[i]

    __intervals = [
        1, 2, 5, 10, 15, 30, 60,
        2 * 60, 10 * 60, 30 * 60, 3600,
        2 * 3600, 3 * 3600, 6 * 3600, 12 * 3600, 24 * 3600,
    ]

    def plot_on_mpl_axes(self, ax):
        """
        Plot this Gantt chart on the provided :class:`matplotlib.axes.Axes`.

        See also :meth:`write_to_png` and :meth:`get_mpl_figure` for the simpler use-cases.
        """
        ordinates = {ident: len(self.__actions) - i for (i, ident) in enumerate(self.__actions.iterkeys())}

        for action in self.__actions.itervalues():
            action.draw(ax, ordinates, self.__actions)

        ax.get_yaxis().set_ticklabels([])
        ax.set_ylim(0.5, len(self.__actions) + 1)

        min_time = min(a.min_time for a in self.__actions.itervalues()).replace(microsecond=0)
        max_time = (
            max(a.max_time for a in self.__actions.itervalues()).replace(microsecond=0) +
            datetime.timedelta(seconds=1)
        )
        duration = int((max_time - min_time).total_seconds())

        ax.set_xlabel("Local time")
        ax.set_xlim(min_time, max_time)
        ax.xaxis_date()
        ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter("%H:%M:%S"))
        ax.xaxis.set_major_locator(matplotlib.dates.AutoDateLocator(maxticks=4, minticks=5))

        ax2 = ax.twiny()
        ax2.set_xlabel("Relative time")
        ax2.set_xlim(min_time, max_time)
        ticks = range(0, duration, self.__nearest(duration // 5, self.__intervals))
        ax2.xaxis.set_ticks([min_time + datetime.timedelta(seconds=s) for s in ticks])
        ax2.xaxis.set_ticklabels(ticks)



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import unittest

from ActionTree import *


class MultipleExecutionsTestCase(unittest.TestCase):
    REPEAT = 5

    def setUp(self):
        self.calls = []

    def __create_mocked_action(self, name):
        mock = unittest.mock.Mock()
        mock.side_effect = lambda: self.calls.append(name)
        action = ActionFromCallable(mock, name)
        return action, mock

    def test_simple_success(self):
        a, aMock = self.__create_mocked_action("a")

        for i in range(self.REPEAT):
            report = execute(a)
            self.assertEqual(report.get_action_status(a).status, ExecutionReport.ActionStatus.Successful)

        call = unittest.mock.call
        self.assertEqual(aMock.mock_calls, [call()] * self.REPEAT)

        self.assertEqual(self.calls, ["a"] * self.REPEAT)

    def test_failure_in_middle(self):
        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        a.add_dependency(b)
        b.add_dependency(c)

        # A lambda that make a side effect then raises an exception
        bMock.side_effect = lambda: [self.calls.append("b"), 1 / 0]

        for i in range(self.REPEAT):
            with self.assertRaises(CompoundException) as catcher:
                execute(a)
            report = catcher.exception.execution_report
            self.assertEqual(report.get_action_status(a).status, ExecutionReport.ActionStatus.Canceled)
            self.assertEqual(report.get_action_status(b).status, ExecutionReport.ActionStatus.Failed)
            self.assertEqual(report.get_action_status(c).status, ExecutionReport.ActionStatus.Successful)

        call = unittest.mock.call
        self.assertEqual(aMock.mock_calls, [])
        self.assertEqual(bMock.mock_calls, [call()] * self.REPEAT)
        self.assertEqual(cMock.mock_calls, [call()] * self.REPEAT)

        self.assertEqual(self.calls, ["c", "b"] * self.REPEAT)



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import unittest

from ActionTree import *


class DependencyCycleTestCase(unittest.TestCase):
    def test_self_dependency(self):
        a = Action("a")
        with self.assertRaises(Exception) as cm:
            a.add_dependency(a)
        self.assertEqual(cm.exception.args[0], "Dependency cycle")

    def test_short_cycle(self):
        a = Action("a")
        b = Action("b")
        a.add_dependency(b)
        with self.assertRaises(Exception) as cm:
            b.add_dependency(a)
        self.assertEqual(cm.exception.args[0], "Dependency cycle")

    def test_long_cycle(self):
        a = Action("a")
        b = Action("b")
        c = Action("c")
        d = Action("d")
        e = Action("e")
        a.add_dependency(b)
        b.add_dependency(c)
        c.add_dependency(d)
        d.add_dependency(e)
        with self.assertRaises(Exception) as cm:
            e.add_dependency(a)
        self.assertEqual(cm.exception.args[0], "Dependency cycle")



#!/usr/bin/env python
# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import setuptools
import sys


version = "0.7.0"


def py2_only(*dependencies):
    if sys.version_info[0] == 2:
        return list(dependencies)
    else:
        return []

setuptools.setup(
    name="ActionTree",
    version=version,
    description="Executes (long) actions in parallel, respecting dependencies between those actions",
    author="Vincent Jacques",
    author_email="vincent@vincent-jacques.net",
    url="http://pythonhosted.org/ActionTree/",
    packages=setuptools.find_packages(),
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development",
    ],
    install_requires=["graphviz", "matplotlib"] + py2_only("futures"),
    tests_require=py2_only("mock"),
    test_suite="ActionTree.tests",
    use_2to3=True,
    command_options={
        "build_sphinx": {
            "version": ("setup.py", version),
            "release": ("setup.py", version),
            "source_dir": ("setup.py", "doc"),
        },
    },
)



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

from __future__ import division, absolute_import, print_function

import time
import unittest

from ActionTree import *


class ExceptionsHandlingTestCase(unittest.TestCase):
    def __create_mocked_action(self, name):
        mock = unittest.mock.Mock()
        action = ActionFromCallable(mock, name)
        return action, mock

    def test_simple_failure(self):
        a, aMock = self.__create_mocked_action("a")

        e = Exception("FooBar")
        aMock.side_effect = e

        with self.assertRaises(CompoundException) as catcher:
            execute(a)
        report = catcher.exception.execution_report

        self.assertEqual(catcher.exception.exceptions, [e])

        aMock.assert_called_once_with()

        self.assertEqual(report.get_action_status(a).status, ExecutionReport.ActionStatus.Failed)

    def test_exception_in_dependency(self):
        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")

        a.add_dependency(b)

        e = Exception()
        bMock.side_effect = e

        with self.assertRaises(CompoundException) as catcher:
            execute(a)
        report = catcher.exception.execution_report

        self.assertEqual(len(catcher.exception.exceptions), 1)
        self.assertIs(catcher.exception.exceptions[0], e)

        bMock.assert_called_once_with()
        aMock.assert_not_called()

        self.assertEqual(report.get_action_status(a).status, ExecutionReport.ActionStatus.Canceled)
        self.assertEqual(report.get_action_status(b).status, ExecutionReport.ActionStatus.Failed)

    def test_exceptions_in_dependencies_with_keep_going(self):
        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        d, dMock = self.__create_mocked_action("d")

        a.add_dependency(b)
        a.add_dependency(c)
        a.add_dependency(d)

        eb = Exception("eb", 42)
        ec = Exception("ec")
        bMock.side_effect = eb
        cMock.side_effect = ec

        with self.assertRaises(CompoundException) as catcher:
            execute(a, keep_going=True)
        report = catcher.exception.execution_report

        self.assertEqual(len(catcher.exception.exceptions), 2)
        self.assertIn(eb, catcher.exception.exceptions)
        self.assertIn(ec, catcher.exception.exceptions)

        aMock.assert_not_called()
        bMock.assert_called_once_with()
        cMock.assert_called_once_with()
        dMock.assert_called_once_with()

        self.assertEqual(report.get_action_status(a).status, ExecutionReport.ActionStatus.Canceled)
        self.assertEqual(report.get_action_status(b).status, ExecutionReport.ActionStatus.Failed)
        self.assertEqual(report.get_action_status(c).status, ExecutionReport.ActionStatus.Failed)
        self.assertEqual(report.get_action_status(d).status, ExecutionReport.ActionStatus.Successful)

    def test_exceptions_in_long_branch_dependencies_with_keep_going(self):
        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        d, dMock = self.__create_mocked_action("d")
        e, eMock = self.__create_mocked_action("e")
        f, fMock = self.__create_mocked_action("f")
        g, gMock = self.__create_mocked_action("g")

        a.add_dependency(b)
        b.add_dependency(c)
        a.add_dependency(d)
        d.add_dependency(e)
        a.add_dependency(f)
        f.add_dependency(g)

        ex = Exception()
        eMock.side_effect = ex

        with self.assertRaises(CompoundException) as catcher:
            execute(a, keep_going=True)
        report = catcher.exception.execution_report

        self.assertEqual(catcher.exception.exceptions, [ex])

        aMock.assert_not_called()
        bMock.assert_called_once_with()
        cMock.assert_called_once_with()
        dMock.assert_not_called()
        eMock.assert_called_once_with()
        fMock.assert_called_once_with()
        gMock.assert_called_once_with()

        self.assertEqual(report.get_action_status(a).status, ExecutionReport.ActionStatus.Canceled)
        self.assertEqual(report.get_action_status(b).status, ExecutionReport.ActionStatus.Successful)
        self.assertEqual(report.get_action_status(c).status, ExecutionReport.ActionStatus.Successful)
        self.assertEqual(report.get_action_status(d).status, ExecutionReport.ActionStatus.Canceled)
        self.assertEqual(report.get_action_status(e).status, ExecutionReport.ActionStatus.Failed)
        self.assertEqual(report.get_action_status(f).status, ExecutionReport.ActionStatus.Successful)
        self.assertEqual(report.get_action_status(g).status, ExecutionReport.ActionStatus.Successful)

    def test_exceptions_in_dependencies_without_keep_going(self):
        a, aMock = self.__create_mocked_action("a")
        b, bMock = self.__create_mocked_action("b")
        c, cMock = self.__create_mocked_action("c")
        d, dMock = self.__create_mocked_action("d")

        a.add_dependency(b)
        a.add_dependency(c)
        a.add_dependency(d)

        def side_effect():
            time.sleep(0.1)
            raise Exception()

        bMock.side_effect = side_effect
        cMock.side_effect = side_effect
        dMock.side_effect = side_effect

        with self.assertRaises(CompoundException) as catcher:
            execute(a)
        report = catcher.exception.execution_report

        failed = len(catcher.exception.exceptions)
        self.assertIn(failed, [1, 2])

        aMock.assert_not_called()
        self.assertEqual(bMock.called + cMock.called + dMock.called, failed)

        self.assertEqual(report.get_action_status(a).status, ExecutionReport.ActionStatus.Canceled)
        self.assertEqual(
            len([x for x in [b, c, d] if report.get_action_status(x).status == ExecutionReport.ActionStatus.Canceled]),
            3 - failed
        )
        self.assertEqual(
            len([x for x in [b, c, d] if report.get_action_status(x).status == ExecutionReport.ActionStatus.Failed]),
            failed
        )
